using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using SimpleJSON;

/// <summary>
/// This plugin is about editing GameObjects (enabling/disabling/deleting) in CUAs.
/// Based on a mix of ideas of CUAshaderMod by Sally Whitemane and DisableCUAObject2 by unique.
/// 
/// Version history:
///   1.0: Initial version
/// </summary>

namespace HAL9001
{
    public class CUA_Editor : MVRScript
    {
        public static readonly string  PLUGIN_VERSION = "1.0";

        public static readonly List<string>  TYPE_OPTIONS = new List<string>  {"GameObjects", "Renderers", "Colliders", "Lights", "Cameras", "Light Probes", "Reflection Probes"};
        public static readonly Color32 COLOUR_ENABLED     = new Color32(192, 255, 192, 255); // green
        public static readonly Color32 COLOUR_DISABLED    = new Color32(255, 255, 192, 255); // yellow
        public static readonly Color32 COLOUR_DISABLED_GO = new Color32(240, 255, 192, 255); // pale green/yellow
        public static readonly Color32 COLOUR_DESTROYED   = new Color32(255, 128, 128, 255); // red


        protected class CUA_GameObject 
        {
            public CUA_Editor         script;
            public string             name;
            public bool               destroyed;
            public GameObject         gameObject;
            public JSONStorableBool   gameObjectToggle;

            public Renderer[]         renderers;
            public JSONStorableBool[] rendererToggles;
            
            public Collider[]         colliders;
            public JSONStorableBool[] colliderToggles;

            public Light              light;

            public Camera             camera;
            public LightProbeGroup    lightProbeGroup;
            public ReflectionProbe    reflectionProbe;

            protected List<JSONStorableBool> m_toggles;
            protected List<UIDynamicToggle>  m_uiElements;


            public CUA_GameObject(CUA_Editor script, Transform transform)
            {
                this.script     = script;
                this.gameObject = transform.gameObject;
                this.destroyed  = false;
                
                name = gameObject.name;
				// some name cleanup
                name = name.Replace("(Clone)", "");
                name = name.Replace("____", "_").Replace("___", "_").Replace("__", "_");

                m_uiElements = new List<UIDynamicToggle>();
                m_toggles      = new List<JSONStorableBool>();

                DetermineComponents();
            }


            public void RestoreFromJSON(JSONClass configuration)
            {
                if (configuration != null)
                {
                    foreach(var toggle in m_toggles)
                    {
                        string value = configuration[toggle.name];
                        if (!string.IsNullOrEmpty(value))
                        {
                            toggle.val = bool.Parse(value);
                            // script.logDebug($"restored {name} = {value}");
                        }
                    }
                }
            }


            public void DetermineComponents()
            {
                if (gameObject != null)
                {
                    int counter;
                    gameObjectToggle = new JSONStorableBool("G_" + name, gameObject.activeSelf, (bool _val) => UpdateGameObject());
                    m_toggles.Add(gameObjectToggle);

                    renderers       = gameObject.GetComponents<Renderer>();
                    rendererToggles = new JSONStorableBool[renderers.Count()];
                    counter         = 0;
                    foreach (Renderer r in renderers)
                    {
                        string prefix = "R";
                        if (renderers.Count() > 1) { prefix += (counter+1); }
                        rendererToggles[counter] = new JSONStorableBool(prefix + "_" + name, r.enabled, (bool _val) => UpdateGameObject());
                        m_toggles.Add(rendererToggles[counter]);
                        counter++;
                    }

                    colliders       = gameObject.GetComponents<Collider>();
                    colliderToggles = new JSONStorableBool[colliders.Count()];
                    counter         = 0;
                    foreach (Collider c in colliders)
                    {
                        string prefix = "C";
                        if (colliders.Count() > 1) { prefix += (counter+1); }
                        colliderToggles[counter] = new JSONStorableBool(prefix + "_" + name, c.enabled, (bool _val) => UpdateGameObject());
                        m_toggles.Add(colliderToggles[counter]);
                        counter++;
                    }

                    light           = gameObject.GetComponent<Light>();
                    camera          = gameObject.GetComponent<Camera>();
                    lightProbeGroup = gameObject.GetComponent<LightProbeGroup>();
                    reflectionProbe = gameObject.GetComponent<ReflectionProbe>();

                    // make the toggles public
                    foreach (var toggle in m_toggles)
                    {
                        script.RegisterBool(toggle);
                    }
                }
            }

            public void DeregisterComponents()
            {
                foreach (var toggle in m_toggles)
                {
                    script.DeregisterBool(toggle);
                }
                gameObjectToggle = null;
                rendererToggles  = null;
                colliderToggles  = null;
                m_toggles.Clear();
            }
                                
            
            public string TypeToString()
            {
                string strOut = "";
                string glue   = "";

                if (renderers.Count() >  0   ) { strOut += glue + renderers.Count() + " Renderer" + (renderers.Count() == 1 ? "" : "s"); glue = ", "; }
                if (colliders.Count() >  0   ) { strOut += glue + colliders.Count() + " Collider" + (colliders.Count() == 1 ? "" : "s"); glue = ", "; }
                if (light             != null) { strOut += glue + "Light";                       glue = ", "; }
                if (camera            != null) { strOut += glue + "Camera";                      glue = ", "; }
                if (lightProbeGroup   != null) { strOut += glue + "LightProbeGroup";             glue = ", "; }
                if (reflectionProbe   != null) { strOut += glue + "ReflectionProbe";             glue = ", "; }
                
                if (glue.Length > 0) { strOut = "[" +strOut + "]"; }

                return strOut;
            }


            public void CreateUI(string filterValue)
            {
                List<UIDynamicToggle> uiElements = new List<UIDynamicToggle>();

                if (   (filterValue == TYPE_OPTIONS[0])                                ||  // game object
                     ( (filterValue == TYPE_OPTIONS[3]) && (light           != null) ) ||  // light
                     ( (filterValue == TYPE_OPTIONS[4]) && (camera          != null) ) ||  // camera
                     ( (filterValue == TYPE_OPTIONS[5]) && (lightProbeGroup != null) ) ||  // light probe
                     ( (filterValue == TYPE_OPTIONS[6]) && (reflectionProbe != null) ) )   // reflection probe
                {
                    // no specific toggle for those > control them via the GameObject
                    UIDynamicToggle tgl = script.CreateToggle(gameObjectToggle, true);
                    tgl.label = name;
                    uiElements.Add(tgl);
                }
                else if (filterValue == TYPE_OPTIONS[1]) // renderers
                {
                    foreach(var rf in rendererToggles)
                    {
                        uiElements.Add(script.CreateToggle(rf, true));
                    }
                }
                else if (filterValue == TYPE_OPTIONS[2]) // colliders
                {
                    foreach(var cf in colliderToggles)
                    {
                        uiElements.Add(script.CreateToggle(cf, true));
                    }
                }

                m_uiElements.AddRange(uiElements);
                UpdateUI();
            }


            public void UpdateGameObject()
            {
				if (!destroyed && script.enabled)
                {
					gameObject.SetActive(gameObjectToggle.val);
					for (int i=0; i < renderers.Count(); i++) { renderers[i].enabled = rendererToggles[i].val; }
					for (int i=0; i < colliders.Count(); i++) { colliders[i].enabled = colliderToggles[i].val; }
				}
				UpdateUI();
            }


            public void UpdateUI()
            {
                foreach (var tgl in m_uiElements) 
                {
                    // set colour based on toggle/destroyed state
                    if (destroyed)            { tgl.backgroundColor = COLOUR_DESTROYED; } // mark destroyed elements specifically
                    else if (tgl.toggle.isOn) { tgl.backgroundColor = COLOUR_ENABLED;   } // otherwise by toggle value
                    else                      { tgl.backgroundColor = COLOUR_DISABLED;  }

                    // special case: element enabled, but game object is not
                    if (!destroyed && tgl.toggle.isOn && !gameObjectToggle.val) { tgl.backgroundColor = COLOUR_DISABLED_GO; }
                }
            }


            public void RestoreGameObject()
            {
				if (!destroyed)
				{					
					gameObject.SetActive(gameObjectToggle.defaultVal);
					for (int i=0; i < renderers.Count(); i++) { renderers[i].enabled = rendererToggles[i].defaultVal; }
					for (int i=0; i < colliders.Count(); i++) { colliders[i].enabled = colliderToggles[i].defaultVal; }
				}
            }


            public void DestroyUI()
            {
                foreach (var uiElement in m_uiElements)
                {
                    script.RemoveSpacer(uiElement);
                }
                m_uiElements.Clear();
            }


            public void EnableElements()
            {
                foreach (var uiElement in m_uiElements) { uiElement.toggle.isOn = true; }
            }


            public void DisableElements()
            {
                foreach (var uiElement in m_uiElements) { uiElement.toggle.isOn = false; }
            }

            public void RestoreElements()
            {
                foreach (var toggle in m_toggles) { toggle.SetValToDefault(); };
            }


            public void DestroyIfDisabled()
            {
                if (!gameObjectToggle.val) 
                { 
                    Destroy(gameObject);
                    destroyed = true;
                    UpdateUI();
                }
            }
        }


        public override void Init()
        {
            try
            {
                m_cuaGameObjects = new List<CUA_GameObject>();
                m_configuration  = null;

                Atom atom = GetContainingAtom();
                if (!atom.type.Equals("CustomUnityAsset"))
                {
                    logError("Atom is not a Custom Unity Asset (CUA).");
                    return;
                }
                SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
                
                m_cuaLoader = atom.GetComponentInChildren<CustomUnityAssetLoader>() ?? GetLoaderFromParentSubscene(atom);
                if (m_cuaLoader == null)
                {
                    logError("Unable to find CustomUnityAssetLoader to register callback to.");
                    return;
                }
                m_cuaLoader.RegisterAssetLoadedCallback(OnCUALoaded);
                
                CreateBaseUI();
            }
            catch (Exception e) { logError($"{e}"); }
        }


        protected CustomUnityAssetLoader GetLoaderFromParentSubscene(Atom atom)
        {
            var parentAtom = atom.parentAtom;
            if (parentAtom == null || !parentAtom.isSubSceneType)
            {
                return null;
            }

            return parentAtom.GetComponentsInChildren<CustomUnityAssetLoader>()
                .Where(loader => loader.containingAtom == atom).FirstOrDefault();
        }


        public void EnableAllElements()
        {
            foreach (CUA_GameObject go in m_cuaGameObjects) { go.EnableElements(); }
        }
        

        public void DisableAllElements()
        {
            foreach (CUA_GameObject go in m_cuaGameObjects) { go.DisableElements(); }
        }


        public void RestoreAllElements()
        {
            foreach (CUA_GameObject go in m_cuaGameObjects) { go.RestoreElements(); }
        }


        public void DestroyDisabledGameObjects()
        {
            foreach (CUA_GameObject go in m_cuaGameObjects) { go.DestroyIfDisabled(); }
        }


        protected void BuildUI(string filterValue)
        {
            foreach (var go in m_cuaGameObjects)
            {
                go.CreateUI(filterValue);
            }
        }
        

        protected void DestroyUI()
        {
            foreach (var go in m_cuaGameObjects)
            {
                go.DestroyUI();
            }
        }


        protected void CreateBaseUI()
        {
            CreateHeader($"CUA Editor v{PLUGIN_VERSION}", false);
                
            var btnRescan = CreateButton("Rescan CUA", false);
			btnRescan.button.onClick.AddListener(RescanCUA);

            m_cuaElementTypes = new JSONStorableStringChooser("typeFilter", TYPE_OPTIONS, TYPE_OPTIONS[0], "Filter", (_val) => ApplyElementFilter(_val));
            m_cuaElementTypes.isStorable   = false;
			m_cuaElementTypes.isRestorable = false;
            CreatePopup(m_cuaElementTypes, true);

            var buttonPrintStructure = CreateButton("Print CUA Structure", false);
            buttonPrintStructure.button.onClick.AddListener(PrintCUA);

            m_destroyOnLoad = new JSONStorableBool("Destroy on Load", false);
			CreateToggle(m_destroyOnLoad, false);
            RegisterBool(m_destroyOnLoad);

            var buttonEnableAll = CreateButton("Enable All Elements", false);
            buttonEnableAll.buttonColor = COLOUR_ENABLED;
            buttonEnableAll.button.onClick.AddListener(EnableAllElements);

            var buttonDisableAll = CreateButton("Disable All Elements", false);
            buttonDisableAll.buttonColor = COLOUR_DISABLED;
            buttonDisableAll.button.onClick.AddListener(DisableAllElements);

            var buttonRestoreAll = CreateButton("Restore All Elements", false);
            buttonRestoreAll.button.onClick.AddListener(RestoreAllElements);

            var buttonExecuteDestroy = CreateButton("Destroy Disabled GameObjects", false);
            buttonExecuteDestroy.buttonColor = COLOUR_DESTROYED;
            buttonExecuteDestroy.button.onClick.AddListener(DestroyDisabledGameObjects);
        }
        

        protected void ApplyElementFilter(string filterValue)
        {
            DestroyUI();
            BuildUI(filterValue);
        }

        
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            // store data for when asset is actually loaded
            m_configuration = jc;
        }
        
       
        protected void OnCUALoaded()
        {
            ScanCUA(); 
            if (m_destroyOnLoad.val)
            {
                DestroyDisabledGameObjects();
            }   
        }


        protected void OnSceneLoaded()
        {
            // nothing here yet            
        }


        public void RescanCUA()
        {
            m_configuration = null; // don't apply anything from before
            ScanCUA();
        }


        protected void ScanCUA()
        {
            if (m_cuaLoader != null)
            {
                DestroyUI();
                DeregisterAndClearGameObjects();

                foreach (Transform child in m_cuaLoader.transform) // this "skips" the scaleBaseNode
                {
                    RecursiveScanCUA(child);
                }

                BuildUI(m_cuaElementTypes.val);
            }
        }


        protected void RecursiveScanCUA(Transform transform)
        {
            if (transform != null)
            {
                CUA_GameObject cuaGameObject = new CUA_GameObject(this, transform);

                m_cuaGameObjects.Add(cuaGameObject);
                foreach (Transform child in transform)
                {
                    RecursiveScanCUA(child);
                }

                cuaGameObject.RestoreFromJSON(m_configuration);
            }
        }


        protected void PrintCUA()
        {
            if (m_cuaLoader != null)
            {
                RecursivePrintCUA(m_cuaLoader.transform, "");
            }
        }


        protected void RecursivePrintCUA(Transform transform, string parentInfo)
        {
            if (transform != null)
            {
                CUA_GameObject cuaGameObject = new CUA_GameObject(this, transform);
                
                string indent = parentInfo.Replace("0", "  ") // no more children
                    .Replace("1", "│")  // vertical line
                    .Replace("2", "├")  // middle child
                    .Replace("3", "└"); // last child

                indent += transform.childCount > 0 ? "┬" : "─"; // only child, or beginning of many?

                logDebug($"{indent}{cuaGameObject.name} {cuaGameObject.TypeToString()}");
                
                parentInfo = parentInfo.Replace("3", "0").Replace("2", "1"); // last child > keep gaps from here onwards, otherwise a vertical bar

                for (int idx = 0 ; idx < transform.childCount ; idx++)
                {
                    string addendum = (idx == transform.childCount-1) ? "3" : "2";
                    RecursivePrintCUA(transform.GetChild(idx), parentInfo + addendum);
                }
            }
        }


        public void DeregisterAndClearGameObjects()
        {
            foreach (var go in m_cuaGameObjects)
            {
                go.DeregisterComponents();
            }
            m_cuaGameObjects.Clear();
        }


        // ********************************************************************
		#region Logging
		
		protected void logDebug(string text) 
		{
			SuperController.LogMessage($"{nameof(CUA_Editor)}: {text}");
		}
		
		protected void logError(string text) 
		{
			SuperController.LogError($"{nameof(CUA_Editor)}: {text}");
		}

		#endregion Logging

	
		// ********************************************************************
		#region Utilities
		
		protected void CreateHeader(string header, bool rightSide, int fieldHeight = 40) 
		{
			JSONStorableString staticDescString      = new JSONStorableString("HDR_" + header, "<size=32><b>" + header + "</b></size>");
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height          = fieldHeight;
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
		}
		
		#endregion Utilities


		// ********************************************************************
		#region Enable, Disable, Destroy
		
		protected void OnEnable()
		{
            foreach (var go in m_cuaGameObjects) { go.UpdateGameObject(); }
		}


		protected void OnDisable() 
		{
            foreach (var go in m_cuaGameObjects) { go.RestoreGameObject(); }
		}


		protected void OnDestroy()
		{
			if (SuperController.singleton != null) 
			{
                SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
			}
		}
		

        #endregion Enable, Disable, Destroy
    
        protected JSONStorableBool       m_destroyOnLoad;
        protected CustomUnityAssetLoader m_cuaLoader;
        protected List<CUA_GameObject>          m_cuaGameObjects;
        protected JSONClass              m_configuration;
        protected JSONStorableStringChooser m_cuaElementTypes;
    }
}
